/*
 * NVRBMTrainer.h
 *
 *  Created on: 24-Jan-2009
 *      Author: spoon
 */

#ifndef NVRBM_H_
#define NVRBM_H_

#include <matrix.h>
#include <nvmatrix.cuh>
#include <thread.h>
#include <queue.h>
#include <WorkRequest.h>
#include <WorkResult.h>
#include "NVRBM_kernel.cuh"

class NVRBMTrainer;

class NVRBM : public Thread {
private:
    // ===========================
    // Host matrices
    // ===========================
    Matrix* hVisHid, *hVisHidInc;
    Matrix* hHidBiases, *hHidBiasInc;
    Matrix* hVisBiases, *hVisBiasInc;
    Matrix* hVisStdevs, *hVisStdevsInc;
    // ---------------------------
    // For PCD
    // ---------------------------
    Matrix* hMiniNegDataPCD;
    // ---------------------------
    // For sparseness
    // ---------------------------
    Matrix* hHidMeans;

    // ===========================
    // Device matrices
    // ===========================
    NVMatrix* data;
    NVMatrix* visHid, *visHidInc;
    NVMatrix* hidBiases, *hidBiasInc;
    NVMatrix* visBiases, *visBiasInc;
    NVMatrix* visStdevs, *visStdevsInc;
    NVMatrix* visHid_T; // view
    NVMatrix* hidMeans;
    // ---------------------------
    // Temporary storage
    // ---------------------------
    NVMatrix* posVisAct, *posHidAct;
    NVMatrix* negVisAct, *negHidAct; // Unnecessary but convenient
    NVMatrix* hidProbs;
    NVMatrix* miniNegData;
    // ---------------------------
    // For PCD
    // ---------------------------
    NVMatrix* miniReconData;
    // ---------------------------
    // For learning vis stdevs
    // ---------------------------
    NVMatrix* miniScaledData;
    NVMatrix* posStdevGrads, *negStdevGrads;
    NVMatrix* stdevGradsTmp1, *stdevGradsTmp2;
    // ---------------------------
    // For sparseness
    // ---------------------------
    NVMatrix* sparseGrads, *sparseGradsSum;

    // ===========================
    // Various learning parameters
    // ===========================
    float epsW, epsVisBias, epsHidBias, epsStdevs;
    float mom, weightCost, visStdev;
    float sparseCost, sparseTarget, sparseDamping;
    int numCD, minibatchSize;
    bool learnStdevs, addGaussianNoise, binaryVis;

    // ===========================
    // RBM dimensions
    // ===========================
    int numVis, numHid;

    int deviceID; // CUDA device to run on

    // ================================
    // Communication with python thread
    // ================================
    Queue<WorkRequest*>* requestQueue;
    Queue<WorkResult*>* resultQueue;

    void initCuda();
    void train();
    void test();
    void syncWithHost();
    void loadNewData(WorkRequest* req);

    void computeHidProbs(NVMatrix* miniData);
    void computeNegData(NVMatrix* out);
    void computeCDWeightGrads(NVMatrix* miniData, bool positivePhase);
    float computeError(NVMatrix* posData, NVMatrix* negData);
    void updateWeights();
protected:
    void* run();
public:
    NVRBM(Matrix* visHid, Matrix* visHidInc, Matrix* hidBiases, Matrix* hidBiasInc,
            Matrix* visBiases, Matrix* visBiasInc, Matrix* visStdevs, Matrix* visStdevsInc,
            Matrix* miniNegDataPCD, Matrix* hidMeans, float epsW, float epsVisBias, float epsHidBias, float epsStdevs,
            float mom, float weightCost, float visStdev, float sparseCost, float sparseTarget,
            float sparseDamping, int numCD, int minibatchSize,
            bool learnStdevs, bool addGaussianNoise, int deviceID);

    virtual ~NVRBM();

    inline Queue<WorkRequest*>* getWorkRequestQueue() {
        return requestQueue;
    }

    inline Queue<WorkResult*>* getWorkResultQueue() {
        return resultQueue;
    }
};

#endif /* NVRBM_H_ */
