import numpy as n
import os
from time import time, asctime, localtime
from numpy.random import randn, rand
from numpy import s_, dot, tile, zeros, ones, zeros_like, array, ones_like
from util import *
from data import *
from options import *
from math import ceil, floor, sqrt
import sys
import shutil
import platform

class GPUModel:
    def __init__(self, model_name, op, load_dic, device_id, filename_options=None):
        # these are input parameters
        self.model_name = model_name
        self.op = op
        self.options = op.options
        self.load_dic = load_dic
        self.device_id = device_id
        self.filename_options = filename_options
        
        self.fill_excused_options()
        #assert self.op.all_values_given()
        
        for o in op.get_options_list():
            setattr(self, o.name, o.value)

        # these are things that the model must remember but they're not input parameters
        
        if load_dic:
            self.model_state = load_dic["model_state"]
            self.save_file = self.options["load_file"].value
            if not os.path.isdir(self.save_file):
                self.save_file = os.path.dirname(self.save_file)
        else:
            self.model_state = {}
            if filename_options is not None:
                self.save_file = model_name + "_" + '_'.join(['%s_%s' % (char, self.options[opt].get_str_value()) for opt, char in filename_options]) + '_' + asctime(localtime()).replace(' ', '_').replace(':', '_')
            self.model_state["train_outputs"] = []
            self.model_state["test_outputs"] = []
            self.model_state["epoch"] = 1
            self.model_state["batchnum"] = self.train_batch_range[0]
        
        self.init_data_providers()

        if load_dic: 
            self.train_data_provider.advance_batch()
            
        self.init_model_state()
        self.setattr_from_model_state()
            
        self.libmodel = __import__('_' + self.model_name)
        self.init_model_lib()
        
    def setattr_from_model_state(self):
        for var, val in self.model_state.iteritems():
            setattr(self, var, val)
            
    def fill_excused_options(self):
        pass
    
    def init_data_providers(self):
        try:
            self.test_data_provider = DataProvider.get_instance(self.data_path, self.test_batch_range, type=self.dp_type)
            self.train_data_provider = DataProvider.get_instance(self.data_path, self.train_batch_range,
                                                                     self.model_state["epoch"], self.model_state["batchnum"],
                                                                     type=self.dp_type)
        except DataProviderException, e:
            print "Unable to create data provider: %s" % e
            self.print_data_providers()
            sys.exit()
        
    def init_model_state(self):
        pass
       
    def init_model_lib(self):
        pass
    
    def train(self):
        print "Training %s" % self.model_name
        self.op.print_values()
        print "========================="
        self.print_model_state()
        print "Running on CUDA device %d" % self.device_id
        print "Current time: %s" % asctime(localtime())
        next_data = self.get_next_batch()
        while self.epoch <= self.num_epochs:
            data = next_data
            self.epoch, self.batchnum = data[0], data[1]
            print "\t%d.%d..." % (self.epoch, self.batchnum),
            sys.stdout.flush()
            
            compute_time_py = time()
            self.start_batch(data)
            
            # load the next batch while the current one is computing
            next_data = self.get_next_batch()
            
            batch_output = self.finish_batch()
            self.train_outputs += [batch_output]
            self.print_train_results()

            if self.get_num_batches_done() % self.testing_freq == 0:
                self.libmodel.syncWithHost()
                test_output = self.get_test_error()
                self.test_outputs += [test_output]
                self.print_test_results()
                self.print_test_status()
                self.conditional_save()
            
            compute_time_py = time() - compute_time_py
            print "(%.3f sec)" % (compute_time_py)
    
    def print_model_state(self):
        pass
    
    def get_num_batches_done(self):
        return len(self.train_batch_range) * (self.epoch - 1) + self.batchnum - self.train_batch_range[0] + 1
    
    def get_next_batch(self, train=True):
        dp = self.train_data_provider
        if not train:
            dp = self.test_data_provider
        return self.parse_batch_data(dp.get_next_batch())
    
    def parse_batch_data(self, batch_data):
        return batch_data[0], batch_data[1], batch_data[2][0]
    
    def start_batch(self, batch_data, test=False):
        self.libmodel.startBatch(batch_data[2], test)
    
    def finish_batch(self):
        return self.libmodel.finishBatch()
    
    def print_train_results(self):
        batch_error = self.train_outputs[-1][0]
        if not (batch_error > 0 and batch_error < 2e20):
            print "NaN! (or something): %f" % batch_error
            sys.exit()

        print "Train error: %.6f " % (batch_error),

    def print_test_results(self):
        batch_error = self.test_outputs[-1][0]
        print "\n\t\tTest error: %.6f" % batch_error,

    def print_test_status(self):
        status = (len(self.test_outputs) == 1 or self.test_outputs[-1][0] < self.test_outputs[-2][0]) and "ok" or "WORSE"
        print status,
        
    def conditional_save(self):
        batch_error = self.test_outputs[-1][0]
        if batch_error > 0 and batch_error < self.max_test_err:
            self.save_state()
        else:
            print "\tTest error > %g, not saving." % self.max_test_err,
    
    def aggregate_test_outputs(self, test_outputs):
        test_error = sum(t[0] for t in test_outputs) / len(self.test_batch_range)
        return (test_error,)
    
    def get_test_error(self):
        first_epoch = self.test_data_provider.curr_epoch
        next_data = self.get_next_batch(train=False)
        test_outputs = []
        while next_data[0] == first_epoch:
            data = next_data
            self.start_batch(data, True)
            next_data = self.get_next_batch(train=False)
            test_outputs += [self.finish_batch()]
        
        return self.aggregate_test_outputs(test_outputs)
            
    def save_state(self):
        for att in self.model_state:
            if hasattr(self, att):
                self.model_state[att] = getattr(self, att)
        
        dic = {"model_state": self.model_state,
               "op": self.op}
            
        checkpoint_dir = os.path.join(self.save_path, self.save_file)
        checkpoint_file = "%d.%d" % (self.epoch, self.batchnum)
        checkpoint_file_full_path = os.path.join(checkpoint_dir, checkpoint_file)
        if not os.path.exists(checkpoint_dir):
            os.makedirs(checkpoint_dir)
    
        pickle(checkpoint_file_full_path, dic)
        
        for f in sorted(os.listdir(checkpoint_dir), key=alphanum_key):
            if sum(os.path.getsize(os.path.join(checkpoint_dir, f2)) for f2 in os.listdir(checkpoint_dir)) > self.max_filesize_mb*1024*1024 and f !=  checkpoint_file:
                os.remove(os.path.join(checkpoint_dir, f))
            else:
                break
            
    @staticmethod
    def load_checkpoint(load_dir):
        if os.path.isdir(load_dir):
            return unpickle(os.path.join(load_dir, sorted(os.listdir(load_dir), key=alphanum_key)[-1]))
        return unpickle(load_dir)
    
    @staticmethod
    def get_default_options_parser():
        op = OptionsParser()
        op.add_option("f", "load_file", StringOptionParser, "Load file", default="", excuses=OptionsParser.EXCLUDE_ALL)
        op.add_option("train-range", "train_batch_range", RangeOptionParser, "Data batch range: training", set_once=True)
        op.add_option("test-range", "test_batch_range", RangeOptionParser, "Data batch range: testing", set_once=True)
        op.add_option("data-provider", "dp_type", StringOptionParser, "Data provider", set_once=True, default="default")
        op.add_option("test-freq", "testing_freq", IntegerOptionParser, "Testing frequency", set_once=True, default=25)
        op.add_option("epochs", "num_epochs", IntegerOptionParser, "Number of epochs", default=500)
        op.add_option("data-path", "data_path", StringOptionParser, "Data path")
        op.add_option("save-path", "save_path", StringOptionParser, "Save path")
        op.add_option("max-filesize", "max_filesize_mb", IntegerOptionParser, "Maximum save file size (MB)", default=5000)
        op.add_option("max-test-err", "max_test_err", FloatOptionParser, "Maximum test error for saving")
        return op

    @staticmethod
    def print_data_providers():
        print "Available data providers:"
        for dp, desc in DataProvider.dp_types.iteritems():
            print "    %s: %s" % (dp, desc)
        print "Note: not all data providers are applicable for every model!"
            
    @staticmethod
    def parse_options(op):
        try:
            load_dic = None
            options = op.parse()
            if options["load_file"].value_given:
                load_dic = GPUModel.load_checkpoint(options["load_file"].value)
                old_op = load_dic["op"]
                old_op.merge_from(op)
                op = old_op
            op.eval_expr_defaults()
            device_id = get_gpu_lock()
            if device_id != GPU_LOCK_NO_LOCK:  
                return op, load_dic, device_id
            print "No free GPUs!"
        except OptionMissingException, e:
            print e
            op.print_usage()
        except OptionException, e:
            print e
        sys.exit()
        
