/*
 * pynvrbm.cu
 *
 *  Created on: 25-Jan-2009
 *      Author: spoon
 */
#include <Python.h>
#include <arrayobject.h>
#include <assert.h>
#include <cutil_inline.h>
#include <cublas.h>
#include <time.h>
#include <matrix.h>
#include <queue.h>
#include <WorkRequest.h>
#include <WorkResult.h>
#include "pynvrbm.cuh"
#include "NVRBM.cuh"

static NVRBM* nvrbm = NULL;

#ifdef EXEC
#include "tests.cu"

#ifdef USE_MKL
#include <mkl.h>
#include <mkl_vsl.h>
#endif /* USE_MKL */
void runDummyRBM() {
    float epsW = 0.0000001, epsVisBias = 0.01, epsHidBias = 0.01, epsStdevs = 0.01;
    float mom = 0.9, weightCost = 0.00005, visStdev = 1;
    float sparseCost = 0, sparseTarget = 0.1, sparseDamping = 0.6;
    int numCD = 1, minibatchSize = 128;

    VSLStreamStatePtr randomStream;
    vslNewStream(&randomStream, VSL_BRNG_MCG31, 7);
    Matrix* data = new Matrix(8192, 3072);
    data->randomizeNormal(randomStream);
    Matrix* visHid_T = new Matrix(9600, 3072);
    Matrix* visHidInc_T = new Matrix(9600, 3072);
    Matrix* visHid = &visHid_T->transpose(true);
    Matrix* visHidInc = &visHidInc_T->transpose(true);
    delete visHid_T;
    delete visHidInc_T;
    visHid->setTrans(true);
    visHidInc->setTrans(true);
    nvrbm = new NVRBM(visHid,
                      visHidInc,
                      new Matrix(1, 9600),
                      new Matrix(1, 9600),
                      new Matrix(1, 3072),
                      new Matrix(1, 3072),
                      new Matrix(1, 3072),
                      new Matrix(1, 3072),
                      NULL,
                      NULL,
                      epsW, epsVisBias, epsHidBias, epsStdevs, mom, weightCost, visStdev,
                      sparseCost, sparseTarget, sparseDamping, numCD, minibatchSize, true);
    nvrbm->start();

    printf("Training GPU RBM on random data.\n");
    nvrbm->getWorkRequestQueue()->enqueue(new WorkRequest(WorkRequest::TRAIN, data));
    BatchDoneResult* res = dynamic_cast<BatchDoneResult*>(nvrbm->getWorkResultQueue()->dequeue());
    double batchError = res->getBatchError();
    double computeTime = res->getComputeTime();
    printf("Train err: %.6f, time: %.3f mins\n", batchError , computeTime / (60*1000));
}

int main (int argc, char* argv) {
//    cudaSetDevice( cutGetMaxGflopsDeviceId() );
//    cublasInit();

    NVMatrix::initDeviceProps();
    NVMatrix::initRandom(time(0));

//    runDummyRBM();
    runMyMultTest();

    return 0;
}
#else
/* ==== Set up the methods table ====================== */
static PyMethodDef _RBMMethods[] = {{ "initNVRBM",      initNVRBM,      METH_VARARGS },
                                    { "startBatch",     startBatch,     METH_VARARGS },
                                    { "finishBatch",    finishBatch,    METH_VARARGS },
                                    { "syncWithHost",   syncWithHost,   METH_VARARGS },
                                    { NULL, NULL } /* Sentinel - marks the end of this structure */
};

/*
 * Module initialization function. Required for python C extensions (and called automatically).
 */
void init_nvrbm() {
    (void) Py_InitModule("_nvrbm", _RBMMethods);
    import_array(); // Must be present for NumPy.  Called first after above line.
}

PyObject* initNVRBM(PyObject *self, PyObject *args) {
    assert(nvrbm == NULL);

    PyArrayObject *visHid, *visBiases, *hidBiases;
    PyArrayObject *visHidInc, *visBiasInc, *hidBiasInc;
    PyArrayObject *visStdevs, *visStdevsInc;
    PyArrayObject *miniNegDataPCD;
    PyArrayObject *hidMeans;
    double visStdev, mom, epsW, epsVisBias, epsHidBias, epsStdevs, weightCost;
    double sparseCost, sparseTarget, sparseDamping;
    int minibatchSize, numCD;
    int learnStdevs, addGaussianNoise;
    int deviceID;

    if (!PyArg_ParseTuple(args, "O!O!O!O!O!O!O!O!O!O!ddddddddddiiiii",
                          &PyArray_Type, &visHid,
                          &PyArray_Type, &visBiases,
                          &PyArray_Type, &hidBiases,
                          &PyArray_Type, &visHidInc,
                          &PyArray_Type, &visBiasInc,
                          &PyArray_Type, &hidBiasInc,
                          &PyArray_Type, &visStdevs,
                          &PyArray_Type, &visStdevsInc,
                          &PyArray_Type, &miniNegDataPCD,
                          &PyArray_Type, &hidMeans,
                          &epsW, &epsVisBias, &epsHidBias, &epsStdevs, &weightCost, &visStdev, &mom,
                          &sparseCost, &sparseTarget, &sparseDamping,
                          &minibatchSize, &numCD, &learnStdevs, &addGaussianNoise, &deviceID)) {
        return NULL;
    }

    // All of these are views. They're never deleted but that's ok.
    nvrbm = new NVRBM(new Matrix(visHid),
                      new Matrix(visHidInc),
                      new Matrix(hidBiases),
                      new Matrix(hidBiasInc),
                      new Matrix(visBiases),
                      new Matrix(visBiasInc),
                      new Matrix(visStdevs),
                      new Matrix(visStdevsInc),
                      new Matrix(miniNegDataPCD),
                      new Matrix(hidMeans),
                      epsW, epsVisBias, epsHidBias, epsStdevs, mom,
                      weightCost, visStdev,
                      sparseCost, sparseTarget, sparseDamping,
                      numCD, minibatchSize, learnStdevs, addGaussianNoise, deviceID);
    nvrbm->start();

    return Py_BuildValue("i", 0);
}

/*
 * Starts training/testing on the given batch (asynchronous -- returns immediately).
 */
PyObject* startBatch(PyObject *self, PyObject *args) {
    assert(nvrbm != NULL);
    PyArrayObject* data;
    int test = 0;

    if (!PyArg_ParseTuple(args, "O!|i",
                          &PyArray_Type, &data,
                          &test)) {
        return NULL;
    }
    nvrbm->getWorkRequestQueue()->enqueue(new WorkRequest(test ? WorkRequest::TEST : WorkRequest::TRAIN, new Matrix(data)));
    return Py_BuildValue("i", 0);
}

/*
 * Waits for the trainer to finish training on the batch given to startBatch.
 * Returns average reconstruction error.
 */
PyObject* finishBatch(PyObject *self, PyObject *args) {
    assert(nvrbm != NULL);
    BatchDoneResult* res = dynamic_cast<BatchDoneResult*>(nvrbm->getWorkResultQueue()->dequeue());

    double batchError = res->getBatchError();
    delete res;
    return Py_BuildValue("(d)", batchError);
}

/*
 * Copies weight matrices from GPU to system memory.
 */
PyObject* syncWithHost(PyObject *self, PyObject *args) {
    assert(nvrbm != NULL);
    nvrbm->getWorkRequestQueue()->enqueue(new WorkRequest(WorkRequest::SYNC));
    SyncDoneResult* res = dynamic_cast<SyncDoneResult*>(nvrbm->getWorkResultQueue()->dequeue());
    delete res;
    return Py_BuildValue("i", 0);
}

#endif
